(* Code for The Interpreter of eXpL : Expression Language *)

signature INTERPRETER = 
sig
        type types
        val table : (string * exp) list ref
        exception runtime_error
        val typeof : exp -> types
        val bigstep : exp -> unit
        val smallstep : exp -> unit
        val interpret : (exp -> unit) -> exp -> unit
end;

structure Interpreter:INTERPRETER =
struct

        datatype types = INT | FLOAT | BOOLEAN | STRING;

        val table = ref [("c",R 299792458.0),("Pi",R Math.pi),("e",R Math.e)];

        exception tycon_mismatch;
        exception undefined_variable of string;
        exception runtime_error;

        local 
        fun asgn s n [] = [(s,n)]
           |asgn s n ((s1,n1)::l) = if(s = s1) then ((s1,n)::l) else ((s1,n1)::(asgn s n l))
        in
        fun assign s n = (table := (asgn s n (!table)); n)
        end;
        
        local 
        fun search s [] = raise (undefined_variable s)
           |search s ((s1,n)::l) = if(s = s1) then n else (search s l)
        in
        fun valueof v = (search v (!table))
        end;

        fun typeof (N _) = INT
           |typeof (R _) = FLOAT
           |typeof (B _) = BOOLEAN
           |typeof (S _) = STRING
           |typeof (V v) = (typeof(valueof v))
           |typeof (PLUS(e1,e2)) = let 
                                        val t1 = typeof e1
                                        val t2 = typeof e2
                                   in
                                        if((t1 = t2) andalso (t1 =INT orelse t1=FLOAT orelse t1=STRING))
                                        then t1 else raise tycon_mismatch
                                   end
           |typeof (MINUS(e1,e2)) = typeof(PLUS(e1,e2))
           |typeof (MULT(e1,e2)) = typeof(PLUS(e1,e2))
           |typeof (DIV(e1,e2)) = typeof(PLUS(e1,e2))
           |typeof (MOD(e1,e2)) =  let 
                                        val t1 = typeof e1
                                        val t2 = typeof e2
                                   in
                                        if((t1 = t2) andalso (t1 =INT)) then t1 else raise tycon_mismatch
                                   end
           |typeof (AND(e1,e2)) =  let 
                                        val t1 = typeof e1
                                        val t2 = typeof e2
                                   in
                                        if((t1 = t2) andalso (t1 =BOOLEAN)) then t1 else raise tycon_mismatch
                                   end
           |typeof (OR(e1,e2)) = typeof(AND(e1,e2))
           |typeof (NOT e) = if ((typeof e)=BOOLEAN) then BOOLEAN else raise tycon_mismatch
           |typeof (IMP(e1,e2)) = typeof(AND(e1,e2))
           |typeof (EQI(e1,e2)) = typeof(AND(e1,e2))
           |typeof (EQ(e1,e2)) = let
                                        val t1 = typeof e1
                                        val t2 = typeof e2
                                 in
                                        if((t1=t2) andalso (t1 <> BOOLEAN)) then BOOLEAN else raise tycon_mismatch
                                 end
           |typeof (LT(e1,e2)) = typeof(EQ(e1,e2))
           |typeof (GT(e1,e2)) = typeof(EQ(e1,e2))
           |typeof (LTE(e1,e2)) = typeof(EQ(e1,e2))
           |typeof (GTE(e1,e2)) = typeof(EQ(e1,e2))
           |typeof (ASMT(s,e)) = (typeof e)
           |typeof (STMT e) = (typeof e)
           |typeof (SEQ(e,_)) = (typeof e)
           ;

        local
        fun intexp (N n) = n
           |intexp (V v) = intexp(valueof(v))
           |intexp (PLUS(e1,e2)) = (intexp e1) + (intexp e2)
           |intexp (MINUS(e1,e2)) = (intexp e1) - (intexp e2)
           |intexp (MULT(e1,e2)) = (intexp e1) * (intexp e2)
           |intexp (DIV(e1,e2)) = (intexp e1) div (intexp e2)
           |intexp (MOD(e1,e2)) = (intexp e1) mod (intexp e2)
           |intexp (ASMT(s,e)) = intexp(assign s (N(intexp e)))
        
        fun realexp (R r) = r
           |realexp (V v) = realexp(valueof(v))
           |realexp (PLUS(e1,e2)) = (realexp e1) + (realexp e2)
           |realexp (MINUS(e1,e2)) = (realexp e1) - (realexp e2)
           |realexp (MULT(e1,e2)) = (realexp e1) * (realexp e2)
           |realexp (DIV(e1,e2)) = (realexp e1) / (realexp e2)
           |realexp (ASMT(s,e)) = realexp(assign s (R(realexp e)))

       fun strexp (S s) = s
          |strexp (V v) = strexp(valueof(v))
          |strexp (PLUS(e1,e2)) = (strexp e1)^(strexp e2)
          |strexp (ASMT(s,e)) = strexp(assign s (S(strexp e)))

        fun boolexp (B b) = b
           |boolexp (V v) = boolexp(valueof(v))
           |boolexp (AND(e1,e2)) = (boolexp e1) andalso (boolexp e2)
           |boolexp (OR(e1,e2)) = (boolexp e1) orelse (boolexp e2)
           |boolexp (NOT e) = not (boolexp e)
           |boolexp (IMP(e1,e2)) = (not (boolexp e1)) orelse (boolexp e2)
           |boolexp (EQI(e1,e2)) = (boolexp e1) = (boolexp e2)
           |boolexp (EQ(e1,e2)) = (case (typeof e1) of INT => ((intexp e1)=(intexp e2))
                                                    | FLOAT => (Real.==(realexp e1,realexp e2))
                                                    | STRING => ((strexp e1)=(strexp e2)))
           |boolexp (LT(e1,e2)) = (case (typeof e1) of INT => ((intexp e1)<(intexp e2))
                                                    | FLOAT => ((realexp e1)<(realexp e2))
                                                    | STRING => ((strexp e1)<(strexp e2)))
           |boolexp (GT(e1,e2)) = not(boolexp(LTE(e1,e2)))
           |boolexp (LTE(e1,e2)) = (case (typeof e1) of INT => ((intexp e1)<=(intexp e2))
                                                    | FLOAT => ((realexp e1)<=(realexp e2))
                                                    | STRING => ((strexp e1)<=(strexp e2)))
           |boolexp (GTE(e1,e2)) = not(boolexp(LT(e1,e2)))
           |boolexp (ASMT(s,e)) = boolexp(assign s (B(boolexp e)))

       fun eval e INT = (Int.toString(intexp e))^" : INT"
          |eval e FLOAT = (Real.toString(realexp e))^" : FLOAT"
          |eval e BOOLEAN = (Bool.toString(boolexp e))^" : BOOLEAN"
          |eval e STRING = (strexp e)^" : STRING"
       in
        fun bigstep e = print((eval e (typeof e))^"\n");
        end;

        local
        fun redint (N a) = (N a)
           |redint (V v) = valueof(v)
           |redint (PLUS(N a,N b)) = N(a+b)
           |redint (PLUS(N a,e)) = PLUS(N a,redint e)
           |redint (PLUS(e1,e2)) = PLUS(redint e1,e2)
           |redint (MINUS(N a,N b)) = N(a-b)
           |redint (MINUS(N a,e)) = MINUS(N a,redint e)
           |redint (MINUS(e1,e2)) = MINUS(redint e1,e2)
           |redint (MULT(N a,N b)) = N(a*b)
           |redint (MULT(N a,e)) = MULT(N a,redint e)
           |redint (MULT(e1,e2)) = MULT(redint e1,e2)
           |redint (DIV(N a,N b)) = N(a div b)
           |redint (DIV(N a,e)) = DIV(N a,redint e)
           |redint (DIV(e1,e2)) = DIV(redint e1,e2)
           |redint (MOD(N a,N b)) = N(a mod b)
           |redint (MOD(N a,e)) = MOD(N a,redint e)
           |redint (MOD(e1,e2)) = MOD(redint e1,e2)
           |redint (ASMT(s,N a)) = (assign s (N a))
           |redint (ASMT(s,e)) = ASMT(s,redint e)

        fun redreal (R a) = (R a)
           |redreal (V v) = valueof(v)
           |redreal (PLUS(R a,R b)) = R(a+b)
           |redreal (PLUS(R a,e)) = PLUS(R a,redreal e)
           |redreal (PLUS(e1,e2)) = PLUS(redreal e1,e2)
           |redreal (MINUS(R a,R b)) = R(a-b)
           |redreal (MINUS(R a,e)) = MINUS(R a,redreal e)
           |redreal (MINUS(e1,e2)) = MINUS(redreal e1,e2)
           |redreal (MULT(R a,R b)) = R(a*b)
           |redreal (MULT(R a,e)) = MULT(R a,redreal e)
           |redreal (MULT(e1,e2)) = MULT(redreal e1,e2)
           |redreal (DIV(R a,R b)) = R(a / b)
           |redreal (DIV(R a,e)) = DIV(R a,redreal e)
           |redreal (DIV(e1,e2)) = DIV(redreal e1,e2)
           |redreal (ASMT(s,R a)) = (assign s (R a))
           |redreal (ASMT(s,e)) = ASMT(s,redreal e)

        fun redstr (S s) = (S s)
           |redstr (V v) = valueof(v)
           |redstr (PLUS(S s1, S s2)) = S(s1^s2)
           |redstr (PLUS(S s1, e)) = PLUS(S s1,redstr e)
           |redstr (PLUS(e1,e2)) = PLUS(redstr e1,e2)
           |redstr (ASMT(s,S s1)) = (assign s (S s1))
           |redstr (ASMT(s,e)) = ASMT(s,redstr e)

        fun redbool (B b) = (B b)
           |redbool (V v) = valueof(v)
           |redbool (AND(B x, B y)) = B(x andalso y)
           |redbool (AND(B x, e)) = AND(B x,redbool e)
           |redbool (AND(e1,e2)) = AND(redbool e1,e2)
           |redbool (OR(B x, B y)) = B(x orelse y)
           |redbool (OR(B x, e)) = OR(B x,redbool e)
           |redbool (OR(e1,e2)) = OR(redbool e1,e2)
           |redbool (NOT(B x)) = B(not x)
           |redbool (NOT e) = NOT(redbool e)
           |redbool (IMP(B x,B y)) = B((not x) orelse y)
           |redbool (IMP(B x, e)) = IMP(B x,redbool e)
           |redbool (IMP(e1,e2)) = IMP(redbool e1,e2)
           |redbool (EQI(B x, B y)) = B(x = y)
           |redbool (EQI(B x, e)) = EQI(B x,redbool e)
           |redbool (EQI(e1,e2)) = EQI(redbool e1,e2)
           |redbool (EQ(N x, N y)) = B(x = y)
           |redbool (EQ(R x, R y)) = B(Real.==(x,y))
           |redbool (EQ(S x, S y)) = B(x = y)
           |redbool (EQ(N x, e)) = EQ(N x,redint e)
           |redbool (EQ(R x, e)) = EQ(R x,redreal e)
           |redbool (EQ(S x, e)) = EQ(S x,redstr e)
           |redbool (EQ(e1,e2)) = (case (typeof e1) of INT => EQ(redint e1,e2)
                                                    | FLOAT => EQ(redreal e1,e2)
                                                    | STRING => EQ(redstr e1,e2))
           |redbool (LT(N x, N y)) = B(x < y)
           |redbool (LT(R x, R y)) = B(x < y)
           |redbool (LT(S x, S y)) = B(x < y)
           |redbool (LT(N x, e)) = LT(N x,redint e)
           |redbool (LT(R x,e)) = LT(R x,redreal e)
           |redbool (LT(S x,e)) = LT(S x,redstr e)
           |redbool (LT(e1,e2)) = (case (typeof e1) of INT => LT(redint e1,e2)
                                                    | FLOAT => LT(redreal e1,e2)
                                                    | STRING => LT(redstr e1,e2))
           |redbool (GT(e1,e2)) = redbool(NOT(LTE(e1,e2)))
           |redbool (LTE(N x, N y)) = B(x <= y)
           |redbool (LTE(R x, R y)) = B(x <= y)
           |redbool (LTE(S x, S y)) = B(x <= y)
           |redbool (LTE(N x, e)) = LTE(N x,redint e)
           |redbool (LTE(R x,e)) = LTE(R x,redreal e)
           |redbool (LTE(S x,e)) = LTE(S x,redreal e)
           |redbool (LTE(e1,e2)) = (case (typeof e1) of INT => LTE(redint e1,e2)
                                                    | FLOAT => LTE(redreal e1,e2)
                                                    | STRING => LTE(redstr e1,e2))
           |redbool (GTE(e1,e2)) = redbool(NOT(LT(e1,e2)))
           |redbool (ASMT(s,B a)) = (assign s (B a))
           |redbool (ASMT(s,e)) = ASMT(s,redbool e)

        fun intexp (N n) = n
           |intexp ie = intexp(redint ie)

        fun realexp (R r) = r
           |realexp re = (realexp(redreal re))

        fun boolexp (B b) = b
           |boolexp be = boolexp(redbool be)

        fun eval e INT = (Int.toString(intexp e))^" : INT"
           |eval e FLOAT = (Real.toString(realexp e))^" : FLOAT"
           |eval e BOOLEAN = (Bool.toString(boolexp e))^" : BOOLEAN"
        in
        fun smallstep e = print((eval e (typeof e))^"\n")
        end;

        local
        fun helper (calculate:(exp->unit)) (STMT e) n = (
                                                        print ("("^(Int.toString n)^")\t");
                                                        (calculate e);
                                                        print("Program exited normally.\n")
                                                     )
           |helper calculate (SEQ(STMT e,l)) n = (
                                                        print("("^(Int.toString n)^")\t");
                                                        (calculate e);
                                                        (helper calculate l (n+1))
                                                 )
        in
        fun interpret c e = (helper c e 1) 
        handle (undefined_variable x) => (print("Variable "^x^" Undefined\nInterpretation Aborted.\n");raise runtime_error)
              | tycon_mismatch => (print ("Type mismatch in expression\nInterpretation Aborted.\n");raise runtime_error)
              | ex => (print("Mathematical error\nInterpretation Aborted.\n");raise runtime_error)
        end;
end;

